@@ -19,34 +19,12 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
-#ifndef USBI_HOTPLUG_H
+#if !defined(USBI_HOTPLUG_H)
 #define USBI_HOTPLUG_H
 
+#ifndef LIBUSBI_H
 #include "libusbi.h"
-
-enum usbi_hotplug_flags {
-	/* This callback is interested in device arrivals */
-	USBI_HOTPLUG_DEVICE_ARRIVED = LIBUSB_HOTPLUG_EVENT_DEVICE_ARRIVED,
-
-	/* This callback is interested in device removals */
-	USBI_HOTPLUG_DEVICE_LEFT = LIBUSB_HOTPLUG_EVENT_DEVICE_LEFT,
-
-	/* IMPORTANT: The values for the below entries must start *after*
-	 * the highest value of the above entries!!!
-	 */
-
-	/* The vendor_id field is valid for matching */
-	USBI_HOTPLUG_VENDOR_ID_VALID = (1U << 3),
-
-	/* The product_id field is valid for matching */
-	USBI_HOTPLUG_PRODUCT_ID_VALID = (1U << 4),
-
-	/* The dev_class field is valid for matching */
-	USBI_HOTPLUG_DEV_CLASS_VALID = (1U << 5),
-
-	/* This callback has been unregistered and needs to be freed */
-	USBI_HOTPLUG_NEEDS_FREE = (1U << 6),
-};
+#endif
 
 /** \ingroup hotplug
  * The hotplug callback structure. The user populates this structure with
@@ -54,17 +32,23 @@ enum usbi_hotplug_flags {
  * to receive notification of hotplug events.
  */
 struct libusb_hotplug_callback {
-	/** Flags that control how this callback behaves */
-	uint8_t flags;
+	/** Context this callback is associated with */
+	struct libusb_context *ctx;
 
-	/** Vendor ID to match (if flags says this is valid) */
-	uint16_t vendor_id;
+	/** Vendor ID to match or LIBUSB_HOTPLUG_MATCH_ANY */
+	int vendor_id;
 
-	/** Product ID to match (if flags says this is valid) */
-	uint16_t product_id;
+	/** Product ID to match or LIBUSB_HOTPLUG_MATCH_ANY */
+	int product_id;
 
-	/** Device class to match (if flags says this is valid) */
-	uint8_t dev_class;
+	/** Device class to match or LIBUSB_HOTPLUG_MATCH_ANY */
+	int dev_class;
+
+	/** Hotplug callback flags */
+	libusb_hotplug_flag flags;
+
+	/** Event(s) that will trigger this callback */
+	libusb_hotplug_event events;
 
 	/** Callback function to invoke for matching event/device */
 	libusb_hotplug_callback_fn cb;
@@ -75,31 +59,24 @@ struct libusb_hotplug_callback {
 	/** User data that will be passed to the callback function */
 	void *user_data;
 
+	/** Callback is marked for deletion */
+	int needs_free;
+
 	/** List this callback is registered in (ctx->hotplug_cbs) */
 	struct list_head list;
 };
 
+typedef struct libusb_hotplug_callback libusb_hotplug_callback;
+
 struct libusb_hotplug_message {
-	/** The hotplug event that occurred */
 	libusb_hotplug_event event;
-
-	/** The device for which this hotplug event occurred */
 	struct libusb_device *device;
-
-	/** List this message is contained in (ctx->hotplug_msgs) */
-	struct list_head list;
 };
 
-#define for_each_hotplug_cb(ctx, c) \
-	for_each_helper(c, &(ctx)->hotplug_cbs, struct libusb_hotplug_callback)
+typedef struct libusb_hotplug_message libusb_hotplug_message;
 
-#define for_each_hotplug_cb_safe(ctx, c, n) \
-	for_each_safe_helper(c, n, &(ctx)->hotplug_cbs, struct libusb_hotplug_callback)
-
-void usbi_hotplug_deregister(struct libusb_context *ctx, int forced);
+void usbi_hotplug_deregister_all(struct libusb_context *ctx);
 void usbi_hotplug_match(struct libusb_context *ctx, struct libusb_device *dev,
-	libusb_hotplug_event event);
-void usbi_hotplug_notification(struct libusb_context *ctx, struct libusb_device *dev,
-	libusb_hotplug_event event);
+			libusb_hotplug_event event);
 
 #endif
