@@ -22,19 +22,32 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
-#include <config.h>
+#include <libusbi.h>
 
 #include <stdint.h>
+#include <errno.h>
 #include <inttypes.h>
 
-#include "libusbi.h"
 #include "wince_usb.h"
 
+// Forward declares
+static int wince_clock_gettime(int clk_id, struct timespec *tp);
+unsigned __stdcall wince_clock_gettime_threaded(void* param);
+
 // Global variables
 uint64_t hires_frequency, hires_ticks_to_ps;
-const uint64_t epoch_time = UINT64_C(116444736000000000);	// 1970.01.01 00:00:000 in MS Filetime
+int errno;
+const uint64_t epoch_time = UINT64_C(116444736000000000);       // 1970.01.01 00:00:000 in MS Filetime
 int windows_version = WINDOWS_CE;
 static int concurrent_usage = -1;
+// Timer thread
+// NB: index 0 is for monotonic and 1 is for the thread exit event
+HANDLE timer_thread = NULL;
+HANDLE timer_mutex = NULL;
+struct timespec timer_tp;
+volatile LONG request_count[2] = {0, 1};	// last one must be > 0
+HANDLE timer_request[2] = { NULL, NULL };
+HANDLE timer_response = NULL;
 HANDLE driver_handle = INVALID_HANDLE_VALUE;
 
 /*
@@ -52,9 +65,9 @@ static char* windows_error_str(uint32_t retval)
 	uint32_t error_code, format_error;
 
 	error_code = retval?retval:GetLastError();
-
+	
 	safe_stprintf(wErr_string, ERR_BUFFER_SIZE, _T("[%d] "), error_code);
-
+	
 	size = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, error_code,
 		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), &wErr_string[safe_tcslen(wErr_string)],
 		ERR_BUFFER_SIZE - (DWORD)safe_tcslen(wErr_string), NULL);
@@ -81,11 +94,11 @@ static char* windows_error_str(uint32_t retval)
 
 static struct wince_device_priv *_device_priv(struct libusb_device *dev)
 {
-	return (struct wince_device_priv *) dev->os_priv;
+        return (struct wince_device_priv *) dev->os_priv;
 }
 
 // ceusbkwrapper to libusb error code mapping
-static int translate_driver_error(int error)
+static int translate_driver_error(int error) 
 {
 	switch (error) {
 		case ERROR_INVALID_PARAMETER:
@@ -100,7 +113,7 @@ static int translate_driver_error(int error)
 		case ERROR_BUSY:
 			return LIBUSB_ERROR_BUSY;
 
-		// Error codes that are either unexpected, or have
+		// Error codes that are either unexpected, or have 
 		// no suitable LIBUSB_ERROR equivilant.
 		case ERROR_CANCELLED:
 		case ERROR_INTERNAL_ERROR:
@@ -136,9 +149,8 @@ static int init_dllimports()
 	return LIBUSB_SUCCESS;
 }
 
-static int init_device(
-	struct libusb_device *dev, UKW_DEVICE drv_dev,
-	unsigned char bus_addr, unsigned char dev_addr)
+static int init_device(struct libusb_device *dev, UKW_DEVICE drv_dev,
+					   unsigned char bus_addr, unsigned char dev_addr)
 {
 	struct wince_device_priv *priv = _device_priv(dev);
 	int r = LIBUSB_SUCCESS;
@@ -156,9 +168,8 @@ static int init_device(
 // Internal API functions
 static int wince_init(struct libusb_context *ctx)
 {
-	int r = LIBUSB_ERROR_OTHER;
+	int i, r = LIBUSB_ERROR_OTHER;
 	HANDLE semaphore;
-	LARGE_INTEGER li_frequency;
 	TCHAR sem_name[11+1+8]; // strlen(libusb_init)+'\0'+(32-bit hex PID)
 
 	_stprintf(sem_name, _T("libusb_init%08X"), (unsigned int)GetCurrentProcessId()&0xFFFFFFFF);
@@ -197,17 +208,37 @@ static int wince_init(struct libusb_context *ctx)
 			goto init_exit;
 		}
 
-		// find out if we have access to a monotonic (hires) timer
-		if (QueryPerformanceFrequency(&li_frequency)) {
-			hires_frequency = li_frequency.QuadPart;
-			// The hires frequency can go as high as 4 GHz, so we'll use a conversion
-			// to picoseconds to compute the tv_nsecs part in clock_gettime
-			hires_ticks_to_ps = UINT64_C(1000000000000) / hires_frequency;
-			usbi_dbg("hires timer available (Frequency: %"PRIu64" Hz)", hires_frequency);
-		} else {
-			usbi_dbg("no hires timer available on this platform");
-			hires_frequency = 0;
-			hires_ticks_to_ps = UINT64_C(0);
+		// Windows CE doesn't have a way of specifying thread affinity, so this code
+		// just has  to hope QueryPerformanceCounter doesn't report different values when
+		// running on different cores.
+		r = LIBUSB_ERROR_NO_MEM;
+		for (i = 0; i < 2; i++) {
+			timer_request[i] = CreateEvent(NULL, TRUE, FALSE, NULL);
+			if (timer_request[i] == NULL) {
+				usbi_err(ctx, "could not create timer request event %d - aborting", i);
+				goto init_exit;
+			}
+		}
+		timer_response = CreateSemaphore(NULL, 0, MAX_TIMER_SEMAPHORES, NULL);
+		if (timer_response == NULL) {
+			usbi_err(ctx, "could not create timer response semaphore - aborting");
+			goto init_exit;
+		}
+		timer_mutex = CreateMutex(NULL, FALSE, NULL);
+		if (timer_mutex == NULL) {
+			usbi_err(ctx, "could not create timer mutex - aborting");
+			goto init_exit;
+		}
+		timer_thread = CreateThread(NULL, 0, wince_clock_gettime_threaded, NULL, 0, NULL);
+		if (timer_thread == NULL) {
+			usbi_err(ctx, "Unable to create timer thread - aborting");
+			goto init_exit;
+		}
+
+		// Wait for timer thread to init before continuing.
+		if (WaitForSingleObject(timer_response, INFINITE) != WAIT_OBJECT_0) {
+			usbi_err(ctx, "Failed to wait for timer thread to become ready - aborting");
+			goto init_exit;
 		}
 	}
 	// At this stage, either we went through full init successfully, or didn't need to
@@ -219,6 +250,30 @@ init_exit: // Holds semaphore here.
 			UkwCloseDriver(driver_handle);
 			driver_handle = INVALID_HANDLE_VALUE;
 		}
+		if (timer_thread) {
+			SetEvent(timer_request[1]); // actually the signal to quit the thread.
+			if (WAIT_OBJECT_0 != WaitForSingleObject(timer_thread, INFINITE)) {
+				usbi_warn(ctx, "could not wait for timer thread to quit");
+				TerminateThread(timer_thread, 1); // shouldn't happen, but we're destroying
+												  // all objects it might have held anyway.
+			}
+			CloseHandle(timer_thread);
+			timer_thread = NULL;
+		}
+		for (i = 0; i < 2; i++) {
+			if (timer_request[i]) {
+				CloseHandle(timer_request[i]);
+				timer_request[i] = NULL;
+			}
+		}
+		if (timer_response) {
+			CloseHandle(timer_response);
+			timer_response = NULL;
+		}
+		if (timer_mutex) {
+			CloseHandle(timer_mutex);
+			timer_mutex = NULL;
+		}
 	}
 
 	if (r != LIBUSB_SUCCESS)
@@ -231,6 +286,7 @@ init_exit: // Holds semaphore here.
 
 static void wince_exit(void)
 {
+	int i;
 	HANDLE semaphore;
 	TCHAR sem_name[11+1+8]; // strlen(libusb_init)+'\0'+(32-bit hex PID)
 
@@ -251,6 +307,29 @@ static void wince_exit(void)
 	if (--concurrent_usage < 0) {	// Last exit
 		exit_polling();
 
+		if (timer_thread) {
+			SetEvent(timer_request[1]); // actually the signal to quit the thread.
+			if (WAIT_OBJECT_0 != WaitForSingleObject(timer_thread, INFINITE)) {
+				usbi_dbg("could not wait for timer thread to quit");
+				TerminateThread(timer_thread, 1);
+			}
+			CloseHandle(timer_thread);
+			timer_thread = NULL;
+		}
+		for (i = 0; i < 2; i++) {
+			if (timer_request[i]) {
+				CloseHandle(timer_request[i]);
+				timer_request[i] = NULL;
+			}
+		}
+		if (timer_response) {
+			CloseHandle(timer_response);
+			timer_response = NULL;
+		}
+		if (timer_mutex) {
+			CloseHandle(timer_mutex);
+			timer_mutex = NULL;
+		}
 		if (driver_handle != INVALID_HANDLE_VALUE) {
 			UkwCloseDriver(driver_handle);
 			driver_handle = INVALID_HANDLE_VALUE;
@@ -293,7 +372,7 @@ static int wince_get_device_list(
 		if (dev) {
 			usbi_dbg("using existing device for %d/%d (session %ld)",
 					bus_addr, dev_addr, session_id);
-			// Release just this element in the device list (as we already hold a
+			// Release just this element in the device list (as we already hold a 
 			// reference to it).
 			UkwReleaseDeviceList(driver_handle, &devices[i], 1);
 			release_list_offset++;
@@ -325,7 +404,7 @@ err_out:
 	*discdevs = new_devices;
 	safe_unref_device(dev);
 	// Release the remainder of the unprocessed device list.
-	// The devices added to new_devices already will still be passed up to libusb,
+	// The devices added to new_devices already will still be passed up to libusb, 
 	// which can dispose of them at its leisure.
 	UkwReleaseDeviceList(driver_handle, &devices[release_list_offset], count - release_list_offset);
 	return r;
@@ -344,8 +423,8 @@ static void wince_close(struct libusb_device_handle *handle)
 }
 
 static int wince_get_device_descriptor(
-	struct libusb_device *device,
-	unsigned char *buffer, int *host_endian)
+   struct libusb_device *device,
+   unsigned char *buffer, int *host_endian)
 {
 	struct wince_device_priv *priv = _device_priv(device);
 
@@ -382,8 +461,8 @@ static int wince_get_config_descriptor(
 }
 
 static int wince_get_configuration(
-	struct libusb_device_handle *handle,
-	int *config)
+   struct libusb_device_handle *handle,
+   int *config)
 {
 	struct wince_device_priv *priv = _device_priv(handle->dev);
 	UCHAR cv = 0;
@@ -499,17 +578,19 @@ static int wince_attach_kernel_driver(
 	struct wince_device_priv *priv = _device_priv(handle->dev);
 	if (!UkwAttachKernelDriver(priv->dev, interface_number)) {
 		return translate_driver_error(GetLastError());
-	}
+	}	
 	return LIBUSB_SUCCESS;
 }
 
-static void wince_destroy_device(struct libusb_device *dev)
+static void wince_destroy_device(
+	struct libusb_device *dev)
 {
 	struct wince_device_priv *priv = _device_priv(dev);
 	UkwReleaseDeviceList(driver_handle, &priv->dev, 1);
 }
 
-static void wince_clear_transfer_priv(struct usbi_transfer *itransfer)
+static void wince_clear_transfer_priv(
+	struct usbi_transfer *itransfer)
 {
 	struct wince_transfer_priv *transfer_priv = (struct wince_transfer_priv*)usbi_transfer_get_os_priv(itransfer);
 	struct winfd wfd = fd_to_winfd(transfer_priv->pollable_fd.fd);
@@ -519,12 +600,13 @@ static void wince_clear_transfer_priv(struct usbi_transfer *itransfer)
 	usbi_free_fd(&transfer_priv->pollable_fd);
 }
 
-static int wince_cancel_transfer(struct usbi_transfer *itransfer)
+static int wince_cancel_transfer(
+	struct usbi_transfer *itransfer)
 {
 	struct libusb_transfer *transfer = USBI_TRANSFER_TO_LIBUSB_TRANSFER(itransfer);
 	struct wince_device_priv *priv = _device_priv(transfer->dev_handle->dev);
 	struct wince_transfer_priv *transfer_priv = (struct wince_transfer_priv*)usbi_transfer_get_os_priv(itransfer);
-
+	
 	if (!UkwCancelTransfer(priv->dev, transfer_priv->pollable_fd.overlapped, UKW_TF_NO_WAIT)) {
 		return translate_driver_error(GetLastError());
 	}
@@ -574,7 +656,7 @@ static int wince_submit_control_or_bulk_transfer(struct usbi_transfer *itransfer
 
 		ret = UkwIssueControlTransfer(priv->dev, flags, setup, buf, bufLen, &transfer->actual_length, wfd.overlapped);
 	} else {
-		ret = UkwIssueBulkTransfer(priv->dev, flags, transfer->endpoint, transfer->buffer,
+		ret = UkwIssueBulkTransfer(priv->dev, flags, transfer->endpoint, transfer->buffer, 
 			transfer->length, &transfer->actual_length, wfd.overlapped);
 	}
 	if (!ret) {
@@ -585,6 +667,7 @@ static int wince_submit_control_or_bulk_transfer(struct usbi_transfer *itransfer
 		return libusbErr;
 	}
 	usbi_add_pollfd(ctx, transfer_priv->pollable_fd.fd, direction_in ? POLLIN : POLLOUT);
+	itransfer->flags |= USBI_TRANSFER_UPDATED_FDS;
 
 	return LIBUSB_SUCCESS;
 }
@@ -594,7 +677,8 @@ static int wince_submit_iso_transfer(struct usbi_transfer *itransfer)
 	return LIBUSB_ERROR_NOT_SUPPORTED;
 }
 
-static int wince_submit_transfer(struct usbi_transfer *itransfer)
+static int wince_submit_transfer(
+	struct usbi_transfer *itransfer)
 {
 	struct libusb_transfer *transfer = USBI_TRANSFER_TO_LIBUSB_TRANSFER(itransfer);
 
@@ -613,9 +697,7 @@ static int wince_submit_transfer(struct usbi_transfer *itransfer)
 	}
 }
 
-static void wince_transfer_callback(
-	struct usbi_transfer *itransfer,
-	uint32_t io_result, uint32_t io_size)
+static void wince_transfer_callback(struct usbi_transfer *itransfer, uint32_t io_result, uint32_t io_size)
 {
 	struct libusb_transfer *transfer = USBI_TRANSFER_TO_LIBUSB_TRANSFER(itransfer);
 	struct wince_transfer_priv *transfer_priv = (struct wince_transfer_priv*)usbi_transfer_get_os_priv(itransfer);
@@ -624,10 +706,10 @@ static void wince_transfer_callback(
 
 	usbi_dbg("handling I/O completion with errcode %d", io_result);
 
-	if (io_result == ERROR_NOT_SUPPORTED &&
+	if (io_result == ERROR_NOT_SUPPORTED && 
 		transfer->type != LIBUSB_TRANSFER_TYPE_CONTROL) {
-		/* For functional stalls, the WinCE USB layer (and therefore the USB Kernel Wrapper
-		 * Driver) will report USB_ERROR_STALL/ERROR_NOT_SUPPORTED in situations where the
+		/* For functional stalls, the WinCE USB layer (and therefore the USB Kernel Wrapper 
+		 * Driver) will report USB_ERROR_STALL/ERROR_NOT_SUPPORTED in situations where the 
 		 * endpoint isn't actually stalled.
 		 *
 		 * One example of this is that some devices will occasionally fail to reply to an IN
@@ -641,7 +723,7 @@ static void wince_transfer_callback(
 		usbi_dbg("checking I/O completion with errcode ERROR_NOT_SUPPORTED is really a stall");
 		if (UkwIsPipeHalted(priv->dev, transfer->endpoint, &halted)) {
 			/* Pipe status retrieved, so now request endpoint status by sending a GET_STATUS
-			 * control request to the device. This is done synchronously, which is a bit
+			 * control request to the device. This is done synchronously, which is a bit 
 			 * naughty, but this is a special corner case.
 			 */
 			WORD wStatus = 0;
@@ -710,9 +792,7 @@ static void wince_transfer_callback(
 	}
 }
 
-static void wince_handle_callback(
-	struct usbi_transfer *itransfer,
-	uint32_t io_result, uint32_t io_size)
+static void wince_handle_callback (struct usbi_transfer *itransfer, uint32_t io_result, uint32_t io_size)
 {
 	struct libusb_transfer *transfer = USBI_TRANSFER_TO_LIBUSB_TRANSFER(itransfer);
 
@@ -724,7 +804,7 @@ static void wince_handle_callback(
 		wince_transfer_callback (itransfer, io_result, io_size);
 		break;
 	case LIBUSB_TRANSFER_TYPE_BULK_STREAM:
-		break;
+		return LIBUSB_ERROR_NOT_SUPPORTED;
 	default:
 		usbi_err(ITRANSFER_CTX(itransfer), "unknown endpoint type %d", transfer->type);
 	}
@@ -773,11 +853,9 @@ static int wince_handle_events(
 			wince_handle_callback(transfer, io_result, io_size);
 		} else if (found) {
 			usbi_err(ctx, "matching transfer for fd %x has not completed", fds[i]);
-			usbi_mutex_unlock(&ctx->open_devs_lock);
 			return LIBUSB_ERROR_OTHER;
 		} else {
 			usbi_err(ctx, "could not find a matching transfer for fd %x", fds[i]);
-			usbi_mutex_unlock(&ctx->open_devs_lock);
 			return LIBUSB_ERROR_NOT_FOUND;
 		}
 	}
@@ -789,20 +867,105 @@ static int wince_handle_events(
 /*
  * Monotonic and real time functions
  */
+unsigned __stdcall wince_clock_gettime_threaded(void* param)
+{
+	LARGE_INTEGER hires_counter, li_frequency;
+	LONG nb_responses;
+	int timer_index;
+
+	// Init - find out if we have access to a monotonic (hires) timer
+	if (!QueryPerformanceFrequency(&li_frequency)) {
+		usbi_dbg("no hires timer available on this platform");
+		hires_frequency = 0;
+		hires_ticks_to_ps = UINT64_C(0);
+	} else {
+		hires_frequency = li_frequency.QuadPart;
+		// The hires frequency can go as high as 4 GHz, so we'll use a conversion
+		// to picoseconds to compute the tv_nsecs part in clock_gettime
+		hires_ticks_to_ps = UINT64_C(1000000000000) / hires_frequency;
+		usbi_dbg("hires timer available (Frequency: %"PRIu64" Hz)", hires_frequency);
+	}
+
+	// Signal wince_init() that we're ready to service requests
+	if (ReleaseSemaphore(timer_response, 1, NULL) == 0) {
+		usbi_dbg("unable to release timer semaphore: %s", windows_error_str(0));
+	}
+
+	// Main loop - wait for requests
+	while (1) {
+		timer_index = WaitForMultipleObjects(2, timer_request, FALSE, INFINITE) - WAIT_OBJECT_0;
+		if ( (timer_index != 0) && (timer_index != 1) ) {
+			usbi_dbg("failure to wait on requests: %s", windows_error_str(0));
+			continue;
+		}
+		if (request_count[timer_index] == 0) {
+			// Request already handled
+			ResetEvent(timer_request[timer_index]);
+			// There's still a possiblity that a thread sends a request between the
+			// time we test request_count[] == 0 and we reset the event, in which case
+			// the request would be ignored. The simple solution to that is to test
+			// request_count again and process requests if non zero.
+			if (request_count[timer_index] == 0)
+				continue;
+		}
+		switch (timer_index) {
+		case 0:
+			WaitForSingleObject(timer_mutex, INFINITE);
+			// Requests to this thread are for hires always
+			if (QueryPerformanceCounter(&hires_counter) != 0) {
+				timer_tp.tv_sec = (long)(hires_counter.QuadPart / hires_frequency);
+				timer_tp.tv_nsec = (long)(((hires_counter.QuadPart % hires_frequency)/1000) * hires_ticks_to_ps);
+			} else {
+				// Fallback to real-time if we can't get monotonic value
+				// Note that real-time clock does not wait on the mutex or this thread.
+				wince_clock_gettime(USBI_CLOCK_REALTIME, &timer_tp);
+			}
+			ReleaseMutex(timer_mutex);
+
+			nb_responses = InterlockedExchange((LONG*)&request_count[0], 0);
+			if ( (nb_responses)
+			  && (ReleaseSemaphore(timer_response, nb_responses, NULL) == 0) ) {
+				usbi_dbg("unable to release timer semaphore: %s", windows_error_str(0));
+			}
+			continue;
+		case 1: // time to quit
+			usbi_dbg("timer thread quitting");
+			return 0;
+		}
+	}
+	usbi_dbg("ERROR: broken timer thread");
+	return 1;
+}
+
 static int wince_clock_gettime(int clk_id, struct timespec *tp)
 {
-	LARGE_INTEGER hires_counter;
-	ULARGE_INTEGER rtime;
 	FILETIME filetime;
+	ULARGE_INTEGER rtime;
+	DWORD r;
 	SYSTEMTIME st;
 	switch(clk_id) {
 	case USBI_CLOCK_MONOTONIC:
-		if (hires_frequency != 0 && QueryPerformanceCounter(&hires_counter)) {
-			tp->tv_sec = (long)(hires_counter.QuadPart / hires_frequency);
-			tp->tv_nsec = (long)(((hires_counter.QuadPart % hires_frequency) / 1000) * hires_ticks_to_ps);
-			return LIBUSB_SUCCESS;
+		if (hires_frequency != 0) {
+			while (1) {
+				InterlockedIncrement((LONG*)&request_count[0]);
+				SetEvent(timer_request[0]);
+				r = WaitForSingleObject(timer_response, TIMER_REQUEST_RETRY_MS);
+				switch(r) {
+				case WAIT_OBJECT_0:
+					WaitForSingleObject(timer_mutex, INFINITE);
+					*tp = timer_tp;
+					ReleaseMutex(timer_mutex);
+					return LIBUSB_SUCCESS;
+				case WAIT_TIMEOUT:
+					usbi_dbg("could not obtain a timer value within reasonable timeframe - too much load?");
+					break; // Retry until successful
+				default:
+					usbi_dbg("WaitForSingleObject failed: %s", windows_error_str(0));
+					return LIBUSB_ERROR_OTHER;
+				}
+			}
 		}
-		// Fall through and return real-time if monotonic read failed or was not detected @ init
+		// Fall through and return real-time if monotonic was not detected @ timer init
 	case USBI_CLOCK_REALTIME:
 		// We follow http://msdn.microsoft.com/en-us/library/ms724928%28VS.85%29.aspx
 		// with a predef epoch_time to have an epoch that starts at 1970.01.01 00:00
@@ -822,48 +985,48 @@ static int wince_clock_gettime(int clk_id, struct timespec *tp)
 }
 
 const struct usbi_os_backend wince_backend = {
-	"Windows CE",
-	0,
-	wince_init,
-	wince_exit,
+        "Windows CE",
+        0,
+        wince_init,
+        wince_exit,
 
-	wince_get_device_list,
+        wince_get_device_list,
 	NULL,				/* hotplug_poll */
-	wince_open,
-	wince_close,
+        wince_open,
+        wince_close,
 
-	wince_get_device_descriptor,
-	wince_get_active_config_descriptor,
-	wince_get_config_descriptor,
+        wince_get_device_descriptor,
+        wince_get_active_config_descriptor,
+        wince_get_config_descriptor,
 	NULL,				/* get_config_descriptor_by_value() */
 
-	wince_get_configuration,
-	wince_set_configuration,
-	wince_claim_interface,
-	wince_release_interface,
+        wince_get_configuration,
+        wince_set_configuration,
+        wince_claim_interface,
+        wince_release_interface,
 
-	wince_set_interface_altsetting,
-	wince_clear_halt,
-	wince_reset_device,
+        wince_set_interface_altsetting,
+        wince_clear_halt,
+        wince_reset_device,
 
 	NULL,				/* alloc_streams */
 	NULL,				/* free_streams */
 
-	wince_kernel_driver_active,
-	wince_detach_kernel_driver,
-	wince_attach_kernel_driver,
+        wince_kernel_driver_active,
+        wince_detach_kernel_driver,
+        wince_attach_kernel_driver,
 
-	wince_destroy_device,
+        wince_destroy_device,
 
-	wince_submit_transfer,
-	wince_cancel_transfer,
-	wince_clear_transfer_priv,
+        wince_submit_transfer,
+        wince_cancel_transfer,
+        wince_clear_transfer_priv,
 
-	wince_handle_events,
-	NULL,				/* handle_transfer_completion() */
+        wince_handle_events,
 
-	wince_clock_gettime,
-	sizeof(struct wince_device_priv),
-	sizeof(struct wince_device_handle_priv),
-	sizeof(struct wince_transfer_priv),
+        wince_clock_gettime,
+        sizeof(struct wince_device_priv),
+        sizeof(struct wince_device_handle_priv),
+        sizeof(struct wince_transfer_priv),
+        0,
 };
